// TSMail v1.3 - Copyright (c) 2000 by Tom Grandgent (tom@grandgent.com)
// Written on 7/18/2000
//
// This can be compiled as a DLL or console-mode EXE without making 
// changes to the source code.

#include <windows.h>
#include <winsock.h>
#include <stdio.h>
#include <io.h>
#include <process.h>

// Do debug logging?
//#define TSM_DEBUG_LOGGING

// Error codes
#define TSM_SUCCESS							0
#define TSM_ERROR_BAD_PARAM					1
#define TSM_ERROR_WINSOCK_INIT_FAILURE		2
#define TSM_ERROR_DNS_FAILURE				3
#define TSM_ERROR_UNABLE_TO_CONNECT			4
#define TSM_ERROR_CONNECTION_RESET			5
#define TSM_ERROR_HELO_NACK					6
#define TSM_ERROR_FILE_OPEN_FAILED			7
#define TSM_ERROR_SENDER_NACK				8
#define TSM_ERROR_RECIPIENT_NACK			9
#define TSM_ERROR_DATA_NACK					10
#define TSM_ERROR_MALLOC_FAILED				11

/* VB Declaration for send_email():

Declare Function send_email Lib "tsmail.dll" (ByVal server As String, _
  ByVal sender As String, ByVal rcpt As String, ByVal subject As String, _
  ByVal text As String, ByVal filename As String, ByVal attachments as String,
  ByVal executeAsync as Long) As Long

 */

// Exported functions
#ifndef _CONSOLE
__declspec(dllexport) 
#endif
int 
#ifndef _CONSOLE
_stdcall 
#endif
send_email(char *server, char *from, char *to, char *subject, char *text, char *filename, char *attachments, int async);

// Private (internal) struct for holding parameters (needed for async executions)
typedef struct email_params
{
	char *server, *from, *to, *subject, *text, *filename, *attachments;
	int async, ret;
} email_params;

// Private functions
void debug_printf(char *text, ...);
int to64(char *inbuf, char *outbuf, int nchars);
void output64chunk(unsigned char c1, unsigned char c2, unsigned char c3, int pads, char *outbuf);
int init_winsock(void);
int recv_to_char(SOCKET *sock, char *buf, int len, char target);
int connect_to_smtp_server(SOCKET *sockptr, char *server);
int send_message(SOCKET sock, char *from, char *to, char *subject, char *body, char *filename, char *attachments);

// Debug logging function
void debug_printf(char *text, ...)
{
#ifdef TSM_DEBUG_LOGGING
	va_list ap;
	char s[32767];
	FILE *fp;

	// Parse the variable arguments and come up with one string
	va_start(ap, text);
	vsprintf(s, text, ap);
	va_end(ap);

	fp = fopen("c:\\tsmail.txt", "a");
	puts(s);
	fprintf(fp, "%s\n", s);
	fclose(fp);
#endif
}

// base64 encoding conversion table
static char basis_64[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/";

// Encode data with base64 encoding
int to64(char *inbuf, char *outbuf, int nchars) 
{
    int ct, ct1, i, bytes_left, limit;
    unsigned char c1, c2, c3;
	limit = 68;

	ct = 0;
	ct1 = 0;

	for(i = 0; i < nchars; i += 3)
	{
		bytes_left = nchars - i;

		c1 = *inbuf++;
		if (bytes_left > 1)
			c2 = *inbuf++;
		if (bytes_left > 2)
			c3 = *inbuf++;

		if (bytes_left == 1)
		{
			output64chunk(c1, 0, 0, 2, outbuf);
			outbuf += 4;
		}
		if (bytes_left == 2)
		{
			output64chunk(c1, c2, 0, 1, outbuf);
			outbuf += 4;
		}
		if (bytes_left > 2)
		{
			output64chunk(c1, c2, c3, 0, outbuf);
			outbuf += 4;
		}
		ct += 4;
		ct1 += 4;
		if (ct1 > limit)
		{ 		  
			*outbuf++ = '\r';
			*outbuf++ = '\n';
			ct1 = 0;
			ct += 2;
		}
	}
	if (ct1 <= limit && ct1 != 0)
	{
		ct += 2;
		*outbuf++ = '\r';
		*outbuf++ = '\n';
	}
	return ct;
}

// Output a chunk of base64-encoded data
void output64chunk(unsigned char c1, unsigned char c2, unsigned char c3, int pads, char *outbuf)
{
    *outbuf++ = basis_64[c1 >> 2];
    *outbuf++ = basis_64[((c1 & 0x03) << 4) | ((c2 & 0xF0) >> 4)];
    if (pads == 2)
	{
        *outbuf++ = '=';
        *outbuf++ = '=';
    }
	else if (pads)
	{
        *outbuf++ = basis_64[((c2 & 0x0F) << 2) | ((c3 & 0xC0) >> 6)];
        *outbuf++ = '=';
    }
	else
	{
        *outbuf++ = basis_64[((c2 & 0x0F) << 2) | ((c3 & 0xC0) >> 6)];
        *outbuf++ = basis_64[c3 & 0x3F];                        
    }  
}

// Initialize Winsock
int init_winsock(void)
{
	WSADATA ws;

	if (WSAStartup(0x101, &ws))
	{
		debug_printf("Winsock initialization failed!");
		return TSM_ERROR_WINSOCK_INIT_FAILURE;
	}

	return 0;
}

// Repeatedly calls recv(), building up a buffer until the specified char is 
// received as part of the incoming data
int recv_to_char(SOCKET *sock, char *buf, int len, char target)
{
	int i, bytesrecvd = 0;

	do
	{
		// Read some data into the buffer
		i = recv(*sock, &buf[bytesrecvd], len - bytesrecvd, 0);
		if (i == SOCKET_ERROR)
		{
			i = WSAGetLastError();
			debug_printf("recv_to_char() Winsock error #%d", i);
			*sock = NULL;
			return bytesrecvd;
		}
		bytesrecvd += i;
		buf[bytesrecvd] = '\0';
		// Scan the buffer for the target character
		for (i = 0; i < bytesrecvd; i++)
			if (buf[i] == target)
				break;
		if (i != bytesrecvd)
			break;
		// Loop to read more data if the buffer has room
	} while (bytesrecvd < len);

	return bytesrecvd;
}

// Connects to the specified SMTP server and says HELO
int connect_to_smtp_server(SOCKET *sockptr, char *server)
{
	SOCKET sock = *sockptr;
	SOCKADDR_IN addr;
	int addrlen = sizeof(SOCKADDR);
	HOSTENT *ht;
	char buf[256];
	char text[256];
	int i;

	// Resolve hostname
	debug_printf("Resolving %s...", server);
	ht = gethostbyname(server);
	if (!ht)
	{
		i = WSAGetLastError();
		switch (i)
		{
			case WSAHOST_NOT_FOUND: sprintf(text, "Unable to resolve '%s':\nAuthoritative Answer Host not found.", server); break;
			default: sprintf(text, "Winsock error #%d while trying to resolve %s!", i, server);
		}
		debug_printf("DNS failure (%d): %s", i, text);
		return TSM_ERROR_DNS_FAILURE;
	}
	memcpy(&(addr.sin_addr), ht->h_addr_list[0], sizeof(addr.sin_addr));

	if (addr.sin_addr.s_addr == INADDR_ANY)
	{
		debug_printf("DNS failure: Unable to resolve %s", server);
		return TSM_ERROR_DNS_FAILURE;
	}

	// Create the socket
	debug_printf("Creating socket...");
	sock = socket(AF_INET, SOCK_STREAM, 0);
	addr.sin_family = PF_INET;
	addr.sin_port   = htons(25);

	// Connect to the remote server
	debug_printf("Connecting to %s on port 25...", inet_ntoa(addr.sin_addr));
	if (connect(sock, (LPSOCKADDR) &addr, sizeof(struct sockaddr)) == SOCKET_ERROR)
	{
		i = WSAGetLastError();
		closesocket(sock);
		sock = INVALID_SOCKET;
		debug_printf("connect() failed with error %d", i);
		return TSM_ERROR_UNABLE_TO_CONNECT;
	}

	// Get welcome message from mail server
	debug_printf("Connected, waiting for welcome message from server...");
	i = recv_to_char(&sock, buf, sizeof(buf), 0x0A);
	if (sock == SOCKET_ERROR)
		return TSM_ERROR_CONNECTION_RESET;
	buf[i] = '\0';
	debug_printf("Received %s", buf);

	// Get local hostname and send HELO
	debug_printf("Getting local hostname...");
	gethostname(text, sizeof(text));
	debug_printf("Sending HELO %s", text);
	sprintf(buf, "HELO %s\r\n", text);
	i = send(sock, buf, strlen(buf), 0);
	if (i == SOCKET_ERROR)
		return TSM_ERROR_CONNECTION_RESET;

	// Get response back from mail server
	debug_printf("Getting response from mail server...");
	i = recv_to_char(&sock, buf, sizeof(buf), 0x0A);
	if (sock == SOCKET_ERROR)
		return TSM_ERROR_CONNECTION_RESET;
	buf[i] = '\0';
	debug_printf("Received %s", buf);
	if (atoi(buf) != 250)
	{
		debug_printf("Negative response to HELO!");
		closesocket(sock);
		return TSM_ERROR_HELO_NACK;
	}

	// Save the socket for the caller
	*sockptr = sock;

	return 0;
}

// Send the actual message
int send_message(SOCKET sock, char *from, char *to, char *subject, char *body, char *filename, char *attachments)
{
	FILE *fp;
	char *filedata = NULL;
	char buf[256];
	char headers[16384];
	int i, j, cnt, bytes_read, filelen, atsent;
	char *tok, *ch;
	char shortfilename[MAX_PATH];
	char *outbound = NULL;
	char inbuf[2000];
	char outbuf[2000];
	char *localatlist = strdup(attachments);

	// See if the user requested a file to be sent
	if (filename && filename[0])
	{
		// Open the file
		fp = fopen(filename, "rb");
		if (!fp)
		{
			closesocket(sock);
			return TSM_ERROR_FILE_OPEN_FAILED;
		}
		// Get its length
		filelen = _filelength(_fileno(fp));
		if (filelen)
		{
			// Allocate memory for the file data
			filedata = (char *) malloc(filelen + 1);
			if (!filedata)
			{
				closesocket(sock);
				return TSM_ERROR_FILE_OPEN_FAILED;
			}
			// Read the file data
			fread(filedata, 1, filelen, fp);
		}
		// Close the file
		fclose(fp);
	}

	// Send MAIL FROM:<>
	debug_printf("Sending MAIL FROM:<%s>", from);
	sprintf(buf, "MAIL FROM:<%s>\r\n", from);
	i = send(sock, buf, strlen(buf), 0);
	if (i == SOCKET_ERROR)
		return TSM_ERROR_CONNECTION_RESET;
	// Get response back from mail server
	debug_printf("Getting response from mail server...");
	i = recv_to_char(&sock, buf, sizeof(buf), 0x0A);
	if (sock == SOCKET_ERROR)
		return TSM_ERROR_CONNECTION_RESET;
	buf[i] = '\0';
	debug_printf("Received %s", buf);
	if (atoi(buf) != 250)
	{
		debug_printf("Negative response to MAIL FROM:<%s>", from);
		closesocket(sock);
		return TSM_ERROR_SENDER_NACK;
	}

	// Send RCPT TO:<>
	debug_printf("Sending RCPT TO:<%s>", to);
	sprintf(buf, "RCPT TO:<%s>\r\n", to);
	i = send(sock, buf, strlen(buf), 0);
	if (i == SOCKET_ERROR)
		return TSM_ERROR_CONNECTION_RESET;
	// Get response back from mail server
	debug_printf("Getting response from mail server...");
	i = recv_to_char(&sock, buf, sizeof(buf), 0x0A);
	if (sock == SOCKET_ERROR)
		return TSM_ERROR_CONNECTION_RESET;
	buf[i] = '\0';
	debug_printf("Received %s", buf);
	if (atoi(buf) != 250)
	{
		debug_printf("Negative response to RCPT TO:<%s>", to);
		closesocket(sock);
		return TSM_ERROR_RECIPIENT_NACK;
	}

	// Send DATA
	debug_printf("Sending DATA");
	sprintf(buf, "DATA\r\n");
	i = send(sock, buf, strlen(buf), 0);
	if (i == SOCKET_ERROR)
		return TSM_ERROR_CONNECTION_RESET;
	// Get response back from mail server
	debug_printf("Getting response from mail server...");
	i = recv_to_char(&sock, buf, sizeof(buf), 0x0A);
	if (sock == SOCKET_ERROR)
		return TSM_ERROR_CONNECTION_RESET;
	buf[i] = '\0';
	debug_printf("Received %s", buf);
	if (atoi(buf) != 354)
	{
		debug_printf("Negative response to DATA");
		closesocket(sock);
		return TSM_ERROR_DATA_NACK;
	}

	// Build headers
	debug_printf("Building from header");
	sprintf(headers, "From: %s\r\n", from);
	debug_printf("Building to header");
	sprintf(&headers[strlen(headers)], "To: %s\r\n", to);
	debug_printf("Building subject header");
	sprintf(&headers[strlen(headers)], "Subject: %s\r\n", subject);
	debug_printf("Building other headers");
	strcat(headers, "X-Mailer: TSMail\r\n");
	if (!localatlist || !localatlist[0])
	{
		debug_printf("No attachments");
		// If no attachments, just a plain text message
		strcat(headers, "Content-Type: text/plain; charset=\"iso-8859-1\"\r\n");
		strcat(headers, "Content-Transfer-Encoding: 7bit\r\n");
		strcat(headers, "\r\n");
	}
	else
	{
		debug_printf("Attachments");
		// Attachments, so make it multipart-MIME
		strcat(headers, "MIME-Version: 1.0\r\n");
		strcat(headers, "Content-Type: multipart/mixed; boundary=\"------------58554A5F21AA\"\r\n");
		strcat(headers, "\r\n");
		strcat(headers, "This is a multi-part message in MIME format.\r\n");
		strcat(headers, "\r\n");
		strcat(headers, "--------------58554A5F21AA\r\n");
		strcat(headers, "Content-Type: text/plain; charset=\"iso-8859-1\"\r\n");
		strcat(headers, "Content-Transfer-Encoding: 7bit\r\n");
		strcat(headers, "\r\n");
	}
	// Send headers
	debug_printf("Sending headers");
	i = send(sock, headers, strlen(headers), 0);
	if (i == SOCKET_ERROR)
		return TSM_ERROR_CONNECTION_RESET;
	// Send message text
	if (body && body[0])
	{
		debug_printf("Sending body");
		i = send(sock, body, strlen(body), 0);
		if (i == SOCKET_ERROR)
			return TSM_ERROR_CONNECTION_RESET;
		// Send an extra two CR/LF's after the body
		i = send(sock, "\r\n\r\n", 4, 0);
		if (i == SOCKET_ERROR)
			return TSM_ERROR_CONNECTION_RESET;
	}
	// Send file
	if (filedata)
	{
		debug_printf("Sending file");
		i = send(sock, filedata, filelen, 0);
		if (i == SOCKET_ERROR)
			return TSM_ERROR_CONNECTION_RESET;
	}
	// Send attachments - loop through the list of attachments using strtok()
	if (localatlist && localatlist[0])
	{
		atsent = 0;
		for (tok = strtok(localatlist, ","); tok; tok = strtok(NULL, ","))
		{
			// Open the file
			fp = fopen(tok, "rb");
			if (fp)
			{
				// Terminate a previous attachment with a MIME part boundary
				if (atsent > 0)
					strcat(outbound, "\r\n--------------58554A5F21AA\r\n");

				// Allocate enough memory to hold the file encoded with base64
				i = (int) ((float) _filelength(_fileno(fp)) * 1.46) + 512;  // compensate for base64 overhead
				outbound = (char *) realloc(outbound, i + (outbound ? strlen(outbound) : 0));
				if (!atsent)
				{
					outbound[0] = '\0';
					strcat(outbound, "\r\n\r\n--------------58554A5F21AA\r\n");
				}
				// Find the filename without path
				strcpy(shortfilename, tok);
				ch = strrchr(shortfilename, '\\');
				if (ch)
					ch++;
				else
					ch = shortfilename;
				// Build the headers for this MIME part
				strcat(outbound, "Content-Type: application/octet-stream; name=\"");
				strcat(outbound, ch);
				strcat(outbound, "\"\r\n");
				strcat(outbound, "Content-Transfer-Encoding: base64\r\n");
				strcat(outbound, "Content-Disposition: attachment; filename=\"");
				strcat(outbound, ch);
				strcat(outbound, "\"\r\n\r\n");
				// Read the file, encoding it into memory with base64
				while (1)
				{
					bytes_read = fread(inbuf, sizeof(char), 1080, fp);
					cnt = to64(inbuf, outbuf, bytes_read);
					j = strlen(outbound);
					memcpy(&outbound[j], outbuf, cnt);
					outbound[j + cnt] = '\0';
					if (feof(fp))
						break;
				}
				// Increment the number of attachments and close the file
				atsent++;
				fclose(fp);
			}
		}
		// If at least one file was attached...
		if (atsent > 0)
		{
			// Terminate the last attachment with a MIME part boundary
			strcat(outbound, "\r\n--------------58554A5F21AA--\r\n\r\n");

			// Send the whole chunk of attachments in one go!
			i = send(sock, outbound, strlen(outbound), 0);
			// Free the memory allocated for attachments
			if (outbound)
			{
				free(outbound);
				outbound = NULL;
			}
			if (i == SOCKET_ERROR)
				return TSM_ERROR_CONNECTION_RESET;
		}
	}
	
	// Send end-of-message marker
	debug_printf("Sending end-of-message marker");
	i = send(sock, "\r\n.\r\n", 5, 0);
	if (i == SOCKET_ERROR)
		return TSM_ERROR_CONNECTION_RESET;
	// Get response back from mail server
	debug_printf("Getting response from mail server...");
	i = recv_to_char(&sock, buf, sizeof(buf), 0x0A);
	if (sock == SOCKET_ERROR)
		return TSM_ERROR_CONNECTION_RESET;
	buf[i] = '\0';
	debug_printf("Received %s", buf);
	if (atoi(buf) != 250)
	{
		debug_printf("Negative response to DATA contents");
		closesocket(sock);
		return TSM_ERROR_DATA_NACK;
	}
	// Be nice and say goodbye, don't be rude now!
	debug_printf("Saying goodbye");
	i = send(sock, "QUIT\r\n", 6, 0);

	// Close the socket, we're all done
	closesocket(sock);

	// Free allocated memory for the file data
	if (filedata)
		free(filedata);

	// Success!
	return 0;
}

// Frees the strings in the given parameters struct
void free_params(email_params *e)
{
	if (e->from)
	{
		free(e->from);
		e->from = NULL;
	}
	if (e->to)
	{
		free(e->to);
		e->from = NULL;
	}
	if (e->subject)
	{
		free(e->subject);
		e->from = NULL;
	}
	if (e->text)
	{
		free(e->text);
		e->from = NULL;
	}
	if (e->filename)
	{
		free(e->filename);
		e->from = NULL;
	}
	if (e->attachments)
	{
		free(e->attachments);
		e->from = NULL;
	}
	if (e->async)
		free(e);
}

// Send an email message - this can be done synchronously or asynchronously
// The parameter is a pointer to a parameters struct
void __cdecl email_worker(void *ptr)
{
	int err;
	SOCKET sock;
	email_params *e = (email_params *) ptr;

	// Initialize Winsock
	e->ret = init_winsock();
	if (e->ret)
	{
		free_params(e);
		return;
	}

	// Connect to the SMTP server
	e->ret = connect_to_smtp_server(&sock, e->server);
	if (e->ret)
	{
		WSACleanup();
		free_params(e);
		return;
	}

	// Send the message
	e->ret = send_message(sock, e->from, e->to, e->subject, e->text, e->filename, e->attachments);
	if (e->ret)
	{
		WSACleanup();
		free_params(e);
		return;
	}

	// Shut down Winsock
	WSACleanup();
	free_params(e);
}

// Send an email message (high level function call, exported)
#ifndef _CONSOLE
__declspec(dllexport) 
#endif
int 
#ifndef _CONSOLE
_stdcall 
#endif
send_email(char *server, char *from, char *to, char *subject, char *text, char *filename, char *attachments, int async)
{
	email_params *e = (email_params *) malloc(sizeof(email_params));
	int ret;

	if (!e)
		return TSM_ERROR_MALLOC_FAILED;

	// Check required parameters
	if (!server || !from || !to || !server[0] || !from[0] || !to[0])
		return TSM_ERROR_BAD_PARAM;

	// Prepare the parameters struct
	memset(e, 0, sizeof(email_params));
	e->async = async;
	if (server)
	{
		e->server = strdup(server);
		if (!e->server)
			return TSM_ERROR_MALLOC_FAILED;
	}
	if (from)
	{
		e->from = strdup(from);
		if (!e->from)
			return TSM_ERROR_MALLOC_FAILED;
	}
	if (to)
	{
		e->to = strdup(to);
		if (!e->to)
			return TSM_ERROR_MALLOC_FAILED;
	}
	if (subject)
	{
		e->subject = strdup(subject);
		if (!e->subject)
			return TSM_ERROR_MALLOC_FAILED;
	}
	if (text)
	{
		e->text = strdup(text);
		if (!e->text)
			return TSM_ERROR_MALLOC_FAILED;
	}
	if (filename)
	{
		e->filename = strdup(filename);
		if (!e->filename)
			return TSM_ERROR_MALLOC_FAILED;
	}
	if (attachments)
	{
		e->attachments = strdup(attachments);
		if (!e->attachments)
			return TSM_ERROR_MALLOC_FAILED;
	}

	// Behave differently based on whether or not asynchronous execution was requested
	if (!async)
	{
		// Synchronous: Send the email and return the result
		email_worker(e);
		ret = e->ret;
		free(e);
		return ret;
	}
	else	// Asynchronous: Send the email in a new thread and return success
	{
		_beginthread(email_worker, 0, e);
		return TSM_SUCCESS;
	}
}

#ifdef _CONSOLE
int main(int argc, char *argv[])
{
	int i;

	printf("Sending mail..\n\n");

	i = send_email(
		"216.47.35.50", 
		"sender@asdfasdf.com", 
		"tgrand@canvaslink.com", 
		"Message from TSMail", 
		"This is a test message from TSMail.  Ah yeah!",
		"k:\\projects\\tsmail\\tsmail.cpp",
		"",
		0);

	printf("Result: %d\n", i);

	//Sleep(10000);	// for async testing

	return 0;
}
#endif